@@ -1,6 +1,6 @@
 # -*- encoding: utf-8 -*-
 """
-keri.core.dbing module
+keri.db.dbing module
 
 
 import lmdb
@@ -47,6 +47,7 @@ l
 import os
 import shutil
 import tempfile
+import stat
 
 from contextlib import contextmanager
 
@@ -57,9 +58,11 @@ try:
 except ImportError:
     import json
 
+from hio.base import doing
 
 from  ..kering import KeriError
 
+
 class DatabaseError(KeriError):
     """
     Database related errors
@@ -67,6 +70,9 @@ class DatabaseError(KeriError):
         raise DatabaseError("error message")
     """
 
+MaxHexDigits =  6
+MaxForks = int("f"*MaxHexDigits, 16)  # 16777215
+
 
 def dgKey(pre, dig):
     """
@@ -101,74 +107,124 @@ def clearDatabaserDir(path):
 
 
 @contextmanager
-def openDatabaser(name="test", cls=None):
+def openLMDB(cls=None, name="test", temp=True, **kwa):
     """
-    Wrapper to enable temporary (test) Databaser instances
-    When used in with statement calls .clearDirPath() on exit of with block
+    Context manager wrapper LMDBer instances.
+    Defaults to temporary databases.
+    Context 'with' statements call .close on exit of 'with' block
 
     Parameters:
-        name is str name of temporary Databaser dirPath  extended name so
-                 can have multiple temporary databasers is use differen name
         cls is Class instance of subclass instance
+        name is str name of LMDBer dirPath so can have multiple databasers
+             at different directory path names thar each use different name
+        temp is Boolean, True means open in temporary directory, clear on close
+                        Otherwise open in persistent directory, do not clear on close
 
     Usage:
 
     with openDatabaser(name="gen1") as baser1:
         baser1.env  ....
 
-    with openDatabaser(name="gen2, cls=Logger)
+    with openDatabaser(name="gen2, cls=Baser)
 
     """
     if cls is None:
-        cls = Databaser
+        cls = LMDBer
     try:
-        databaser = cls(name=name, temp=True)
-
-        yield databaser
+        lmdber = cls(name=name, temp=temp, reopen=True, **kwa)
+        yield lmdber
 
     finally:
-
-        databaser.clearDirPath()
+        lmdber.close()
 
 
-class Databaser:
+class LMDBer:
     """
-    Databaser base class for LMDB instances.
+    LBDBer base class for LMDB manager instances.
     Creates a specific instance of an LMDB database directory and environment.
 
     Attributes:
         .name is LMDB database name did2offer
-        .env is LMDB main (super) database environment
+        .temp is Boolean, True means open db in /tmp directory
+        .headDirPath is head directory path for db
+        .mode is numeric os dir permissions for db directory
         .path is LMDB main (super) database directory path
+        .env is LMDB main (super) database environment
+        .opened is Boolean, True means LMDB .env at .path is opened.
+                            Otherwise LMDB .env is closed
 
     Properties:
 
 
     """
-    HeadDirPath = "/var"  # default in /var
+    HeadDirPath = "/usr/local/var"  # default in /usr/local/var
     TailDirPath = "keri/db"
-    AltHeadDirPath = "~"  #  put in ~ when /var not permitted
+    AltHeadDirPath = "~"  #  put in ~ as fallback when desired not permitted
     AltTailDirPath = ".keri/db"
+    TempHeadDir = "/tmp"
+    TempPrefix = "keri_lmdb_"
+    TempSuffix = "_test"
     MaxNamedDBs = 16
 
-    def __init__(self, headDirPath=None, name='main', temp=False):
+    def __init__(self, name='main', temp=False, headDirPath=None, dirMode=None,
+                 reopen=True):
         """
         Setup main database directory at .dirpath.
         Create main database environment at .env using .dirpath.
 
         Parameters:
-            headDirPath is str head of the pathname of directory for main database
-                If not provided use default headDirpath
-            name is str pathname differentiator for directory for main database
-                When system employs more than one keri databse name allows
+            name is str directory path name differentiator for main database
+                When system employs more than one keri database, name allows
                 differentiating each instance by name
-            temp is boolean If True then use temporary head pathname  instead of
-                headDirPath if any or default headDirPath
+            temp is boolean, assign to .temp
+                True then open in temporary directory, clear on close
+                Othewise then open persistent directory, do not clear on close
+            headDirPath is optional str head directory pathname for main database
+                If not provided use default .HeadDirpath
+            dirMode is int numeric os dir permissions for database directory
+                default is use os defaults and not set the dirMode
+            reopen is boolean, IF True then database will be reopened by this init
+
         """
         self.name = name
+        self.temp = True if temp else False
+        self.headDirPath = headDirPath
+        self.dirMode = dirMode
+        self.path = None
+        self.env = None
+        self.opened = False
 
-        if temp:
-            headDirPath = tempfile.mkdtemp(prefix="keri_lmdb_", suffix="_test", dir="/tmp")
+        if reopen:
+            self.reopen(headDirPath=self.headDirPath, dirMode=dirMode)
+
+
+    def reopen(self, temp=None, headDirPath=None, dirMode=None):
+        """
+        Use or Create if not preexistent, directory path for lmdb at .path
+        Open lmdb and assign to .env
+
+        Parameters:
+            temp is optional boolean:
+                If None ignore Otherwise
+                    Assign to .temp
+                    If True then open temporary directory, clear on close
+                    If False then open persistent directory, do not clear on close
+            headDirPath is optional str head directory pathname of main database
+                If not provided use default .HeadDirpath
+        """
+        if temp is not None:
+            self.temp = True if temp else False  # need .temp for clear on .close
+
+        if headDirPath is None:
+            headDirPath = self.headDirPath
+
+        if dirMode is None:
+            dirMode = self.dirMode
+
+        if self.temp:
+            headDirPath = tempfile.mkdtemp(prefix=self.TempPrefix,
+                                           suffix=self.TempSuffix,
+                                           dir=self.TempHeadDir)
             self.path = os.path.abspath(
                                 os.path.join(headDirPath,
                                              self.TailDirPath,
@@ -208,14 +264,20 @@ class Databaser:
                     if not os.path.exists(self.path):
                         os.makedirs(self.path)
 
+            if dirMode is not None:  # set mode if mode and not temp
+                os.chmod(self.path, dirMode)
+
         # open lmdb major database instance
         # creates files data.mdb and lock.mdb in .dbDirPath
         self.env = lmdb.open(self.path, max_dbs=self.MaxNamedDBs)
+        self.opened = True
 
 
-    def clearDirPath(self):
+    def close(self, clear=False):
         """
-        Remove .dirPath
+        Close lmdb at .env and if clear or .temp then remove lmdb directory at .path
+        Parameters:
+           clear is boolean, True means clear lmdb directory
         """
         if self.env:
             try:
@@ -223,6 +285,17 @@ class Databaser:
             except:
                 pass
 
+        self.env = None
+        self.opened = False
+
+        if clear or self.temp:
+            self.clearDirPath()
+
+
+    def clearDirPath(self):
+        """
+        Remove lmdb directory at .path
+        """
         if os.path.exists(self.path):
             shutil.rmtree(self.path)
 
@@ -388,6 +461,7 @@ class Databaser:
                 count = cursor.count()
             return count
 
+
     def delVals(self,db, key, dupdata=True):
         """
         Deletes all values at key in db.
@@ -432,6 +506,9 @@ class Databaser:
             result = False
             for val in vals:
                 if val not in dups:
+                    if cnt > MaxForks:
+                        raise DatabaseError("Too many recovery forks at key = "
+                                            "{}.".format(key))
                     result = True
                     val = (b'%06x.' % (cnt)) +  val  # prepend ordering prefix
                     txn.put(key, val, dupdata=True)
@@ -445,7 +522,7 @@ class Databaser:
         Adds to existing values at key if any
         Returns True if written else False if val is already a dup
 
-        DDuplicates preserve insertion order.
+        Duplicates preserve insertion order.
 
         Parameters:
             db is opened named sub db with dupsort=False
@@ -460,6 +537,9 @@ class Databaser:
                 cnt = cursor.count()
             result = False
             if val not in dups:
+                if cnt > MaxForks:
+                    raise DatabaseError("Too many recovery forks at key = "
+                                        "{}.".format(key))
                 val = (b'%06x.' % (cnt)) +  val  # prepend ordering prefix
                 result = txn.put(key, val, dupdata=True)
             return result
@@ -652,72 +732,104 @@ class Databaser:
 
 
 
-def openLogger(name="test"):
+def openDB(name="test", **kwa):
     """
-    Returns contextmanager generated by openDatabaser but with Logger instance
+    Returns contextmanager generated by openLMDB but with Baser instance
     """
-    return openDatabaser(name=name, cls=Logger)
+    return openLMDB(cls=Baser, name=name, **kwa)
 
 
-class Logger(Databaser):
+class Baser(LMDBer):
     """
-    Logger sets up named sub databases with Keri Event Logs within main database
+    Baser sets up named sub databases with Keri Event Logs within main database
 
     Attributes:
-        see superclass Databaser for inherited attributes
+        see superclass LMDBer for inherited attributes
 
         .evts is named sub DB whose values are serialized events
+            dgKey
             DB is keyed by identifer prefix plus digest of serialized event
             Only one value per DB key is allowed
 
         .dtss is named sub DB of datetime stamp strings in ISO 8601 format of
+            dgKey
             the datetime when the event was first seen by log.
             Used for escrows timeouts and extended validation.
             DB is keyed by identifer prefix plus digest of serialized event
 
         .sigs is named sub DB of fully qualified event signatures
+            dgKey
             DB is keyed by identifer prefix plus digest of serialized event
             More than one value per DB key is allowed
 
-        .rcts is named sub DB of event receipt couplets. Each couplet is
-            concatenation of fully qualified witness or validator prefix plus
-            fully qualified event signature by witness or validator
+        .rcts is named sub DB of event receipt couplets from nontransferable
+            signers. Each couplet is concatenation of fully qualified
+            non-transferale prefix plus fully qualified event signature
+            by witness, watcher, or validator.
+            dgKey
             SB is keyed by identifer prefix plus digest of serialized event
             More than one value per DB key is allowed
 
-        .ures is named sub DB of unverified event receipt escrowed couplets.
-            Each couplet is concatenation of fully qualified identfier prefix
-            for corresponding event plus fully qualified event signature
-            by witness or validator
-            SB is keyed by witness or validator prefix plus digest of serialized event
-            Only one value per DB key is allowed
+        .ures is named sub DB of unverified event receipt escrowed couplets from
+            non-transferable signers. Each couplet is concatenation of fully
+            qualified non-transferable identfier prefix plus fully qualified
+            event signature by witness, watcher, or validator
+            dgKey
+            SB is keyed by controller prefix plus digest
+            of serialized event
+            More than one value per DB key is allowed
+
+        .vrcs is named sub DB of event validator receipt triplets from transferable
+            signers. Each triplet is concatenation of  three fully qualified items
+            of validator. These are transferable prefix, plus latest establishment
+            event digest, plus event signature.
+            When latest establishment event is multisig then there will
+            be multiple triplets one per signing key, each a dup at same db key.
+            dgKey
+            SB is keyed by identifer prefix plus digest of serialized event
+            More than one value per DB key is allowed
+
+        .vres is named sub DB of unverified event validator receipt escrowed
+            triplets from transferable signers. Each triplet is concatenation of
+            three fully qualified items of validator. These are transferable
+            prefix, plus latest establishment event digest, plus event signature.
+            When latest establishment event is multisig then there will
+            be multiple triplets one per signing key, each a dup at same db key.
+            dgKey
+            SB is keyed by identifer prefix plus digest of serialized event
+            More than one value per DB key is allowed
 
         .kels is named sub DB of key event log tables that map sequence numbers
             to serialized event digests.
+            snKey
             Values are digests used to lookup event in .evts sub DB
             DB is keyed by identifer prefix plus sequence number of key event
             More than one value per DB key is allowed
 
         .pses is named sub DB of partially signed escrowed event tables
             that map sequence numbers to serialized event digests.
+            snKey
             Values are digests used to lookup event in .evts sub DB
             DB is keyed by identifer prefix plus sequence number of key event
             More than one value per DB key is allowed
 
         .ooes is named sub DB of out of order escrowed event tables
             that map sequence numbers to serialized event digests.
+            snKey
             Values are digests used to lookup event in .evts sub DB
             DB is keyed by identifer prefix plus sequence number of key event
             More than one value per DB key is allowed
 
         .dels is named sub DB of deplicitous event log tables that map sequence numbers
             to serialized event digests.
+            snKey
             Values are digests used to lookup event in .evts sub DB
             DB is keyed by identifer prefix plus sequence number of key event
             More than one value per DB key is allowed
 
         .ldes is named sub DB of likely deplicitous escrowed event tables
             that map sequence numbers to serialized event digests.
+            snKey
             Values are digests used to lookup event in .evts sub DB
             DB is keyed by identifer prefix plus sequence number of key event
             More than one value per DB key is allowed
@@ -727,11 +839,21 @@ class Logger(Databaser):
 
 
     """
-    def __init__(self, **kwa):
+    def __init__(self, headDirPath=None, reopen=True, **kwa):
         """
         Setup named sub databases.
 
-        Parameters:
+        Inherited Parameters:
+            name is str directory path name differentiator for main database
+                When system employs more than one keri database, name allows
+                differentiating each instance by name
+            temp is boolean, assign to .temp
+                True then open in temporary directory, clear on close
+                Othewise then open persistent directory, do not clear on close
+            headDirPath is optional str head directory pathname for main database
+                If not provided use default .HeadDirpath
+            mode is int numeric os dir permissions for database directory
+            reopen is boolean, IF True then database will be reopened by this init
 
         Notes:
 
@@ -745,7 +867,14 @@ class Logger(Databaser):
         Duplicates are inserted in lexocographic order by value, insertion order.
 
         """
-        super(Logger, self).__init__(**kwa)
+        super(Baser, self).__init__(headDirPath=headDirPath, reopen=reopen, **kwa)
+
+
+    def reopen(self, **kwa):
+        """
+        Open sub databases
+        """
+        super(Baser, self).reopen(**kwa)
 
         # Create by opening first time named sub DBs within main DB instance
         # Names end with "." as sub DB name must include a non Base64 character
@@ -755,7 +884,9 @@ class Logger(Databaser):
         self.dtss = self.env.open_db(key=b'dtss.')
         self.sigs = self.env.open_db(key=b'sigs.', dupsort=True)
         self.rcts = self.env.open_db(key=b'rcts.', dupsort=True)
-        self.ures = self.env.open_db(key=b'ures.')
+        self.ures = self.env.open_db(key=b'ures.', dupsort=True)
+        self.vrcs = self.env.open_db(key=b'vrcs.', dupsort=True)
+        self.vres = self.env.open_db(key=b'vres.', dupsort=True)
         self.kels = self.env.open_db(key=b'kels.', dupsort=True)
         self.pses = self.env.open_db(key=b'pses.', dupsort=True)
         self.ooes = self.env.open_db(key=b'ooes.', dupsort=True)
@@ -765,6 +896,7 @@ class Logger(Databaser):
 
     def putEvt(self, key, val):
         """
+        Use dgKey()
         Write serialized event bytes val to key
         Does not overwrite existing val if any
         Returns True If val successfully written Else False
@@ -774,6 +906,7 @@ class Logger(Databaser):
 
     def setEvt(self, key, val):
         """
+        Use dgKey()
         Write serialized event bytes val to key
         Overwrites existing val if any
         Returns True If val successfully written Else False
@@ -782,6 +915,7 @@ class Logger(Databaser):
 
     def getEvt(self, key):
         """
+        Use dgKey()
         Return event at key
         Returns None if no entry at key
         """
@@ -790,6 +924,7 @@ class Logger(Databaser):
 
     def delEvt(self, key):
         """
+        Use dgKey()
         Deletes value at key.
         Returns True If key exists in database Else False
         """
@@ -798,6 +933,7 @@ class Logger(Databaser):
 
     def putDts(self, key, val):
         """
+        Use dgKey()
         Write serialized event datetime stamp val to key
         Does not overwrite existing val if any
         Returns True If val successfully written Else False
@@ -808,6 +944,7 @@ class Logger(Databaser):
 
     def setDts(self, key, val):
         """
+        Use dgKey()
         Write serialized event datetime stamp val to key
         Overwrites existing val if any
         Returns True If val successfully written Else False
@@ -817,6 +954,7 @@ class Logger(Databaser):
 
     def getDts(self, key):
         """
+        Use dgKey()
         Return datetime stamp at key
         Returns None if no entry at key
         """
@@ -825,6 +963,7 @@ class Logger(Databaser):
 
     def delDts(self, key):
         """
+        Use dgKey()
         Deletes value at key.
         Returns True If key exists in database Else False
         """
@@ -833,6 +972,7 @@ class Logger(Databaser):
 
     def getSigs(self, key):
         """
+        Use dgKey()
         Return list of signatures at key
         Returns empty list if no entry at key
         Duplicates are retrieved in lexocographic order not insertion order.
@@ -842,6 +982,7 @@ class Logger(Databaser):
 
     def getSigsIter(self, key):
         """
+        Use dgKey()
         Return iterator of signatures at key
         Raises StopIteration Error when empty
         Duplicates are retrieved in lexocographic order not insertion order.
@@ -851,6 +992,7 @@ class Logger(Databaser):
 
     def putSigs(self, key, vals):
         """
+        Use dgKey()
         Write each entry from list of bytes signatures vals to key
         Adds to existing signatures at key if any
         Returns True If no error
@@ -862,6 +1004,7 @@ class Logger(Databaser):
 
     def addSig(self, key, val):
         """
+        Use dgKey()
         Add signature val bytes as dup to key in db
         Adds to existing values at key if any
         Returns True if written else False if dup val already exists
@@ -872,6 +1015,7 @@ class Logger(Databaser):
 
     def getSigs(self, key):
         """
+        Use dgKey()
         Return list of signatures at key
         Returns empty list if no entry at key
         Duplicates are retrieved in lexocographic order not insertion order.
@@ -881,6 +1025,7 @@ class Logger(Databaser):
 
     def cntSigs(self, key):
         """
+        Use dgKey()
         Return count of signatures at key
         Returns zero if no entry at key
         """
@@ -889,6 +1034,7 @@ class Logger(Databaser):
 
     def delSigs(self, key):
         """
+        Use dgKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
@@ -897,6 +1043,7 @@ class Logger(Databaser):
 
     def putRcts(self, key, vals):
         """
+        Use dgKey()
         Write each entry from list of bytes receipt couplets vals to key
         Adds to existing receipts at key if any
         Returns True If no error
@@ -908,6 +1055,7 @@ class Logger(Databaser):
 
     def addRct(self, key, val):
         """
+        Use dgKey()
         Add receipt couplet val bytes as dup to key in db
         Adds to existing values at key if any
         Returns True if written else False if dup val already exists
@@ -918,6 +1066,7 @@ class Logger(Databaser):
 
     def getRcts(self, key):
         """
+        Use dgKey()
         Return list of receipt couplets at key
         Returns empty list if no entry at key
         Duplicates are retrieved in lexocographic order not insertion order.
@@ -927,6 +1076,7 @@ class Logger(Databaser):
 
     def getRctsIter(self, key):
         """
+        Use dgKey()
         Return iterator of receipt couplets at key
         Raises StopIteration Error when empty
         Duplicates are retrieved in lexocographic order not insertion order.
@@ -936,6 +1086,7 @@ class Logger(Databaser):
 
     def cntRcts(self, key):
         """
+        Use dgKey()
         Return count of receipt couplets at key
         Returns zero if no entry at key
         """
@@ -944,49 +1095,199 @@ class Logger(Databaser):
 
     def delRcts(self, key):
         """
+        Use dgKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
         return self.delVals(self.rcts, key)
 
 
-    def putUre(self, key, val):
+    def putUres(self, key, vals):
         """
-        Write prefix plus signature couplet val to key
-        Does not overwrite existing val if any
-        Returns True If val successfully written Else False
-        Returns False if key already exists
+        Use dgKey()
+        Write each entry from list of bytes receipt couplets vals to key
+        Adds to existing receipts at key if any
+        Returns True If no error
+        Apparently always returns True (is this how .put works with dupsort=True)
+        Duplicates are inserted in lexocographic order not insertion order.
         """
-        return self.putVal(self.ures, key, val)
+        return self.putVals(self.ures, key, vals)
 
 
-    def setUre(self, key, val):
+    def addUre(self, key, val):
         """
-        Write prefix plus signature couplet val to key
-        Overwrites existing val if any
-        Returns True If val successfully written Else False
+        Use dgKey()
+        Add receipt couplet val bytes as dup to key in db
+        Adds to existing values at key if any
+        Returns True if written else False if dup val already exists
+        Duplicates are inserted in lexocographic order not insertion order.
         """
-        return self.setVal(self.ures, key, val)
+        return self.addVal(self.ures, key, val)
 
 
-    def getUre(self, key):
+    def getUres(self, key):
         """
-        Return prefix plus signature couplet val at key
-        Returns None if no entry at key
+        Use dgKey()
+        Return list of receipt couplets at key
+        Returns empty list if no entry at key
+        Duplicates are retrieved in lexocographic order not insertion order.
         """
-        return self.getVal(self.ures, key)
+        return self.getVals(self.ures, key)
 
 
-    def delUre(self, key):
+    def getUresIter(self, key):
         """
-        Deletes value at key.
+        Use dgKey()
+        Return iterator of receipt couplets at key
+        Raises StopIteration Error when empty
+        Duplicates are retrieved in lexocographic order not insertion order.
+        """
+        return self.getValsIter(self.ures, key)
+
+
+    def cntUres(self, key):
+        """
+        Use dgKey()
+        Return count of receipt couplets at key
+        Returns zero if no entry at key
+        """
+        return self.cntVals(self.ures, key)
+
+
+    def delUres(self, key):
+        """
+        Use dgKey()
+        Deletes all values at key.
         Returns True If key exists in database Else False
         """
-        return self.delVal(self.ures, key)
+        return self.delVals(self.ures, key)
+
+
+    def putVrcs(self, key, vals):
+        """
+        Use dgKey()
+        Write each entry from list of bytes receipt triplets vals to key
+        Adds to existing receipts at key if any
+        Returns True If no error
+        Apparently always returns True (is this how .put works with dupsort=True)
+        Duplicates are inserted in lexocographic order not insertion order.
+        """
+        return self.putVals(self.vrcs, key, vals)
+
+
+    def addVrc(self, key, val):
+        """
+        Use dgKey()
+        Add receipt triplet val bytes as dup to key in db
+        Adds to existing values at key if any
+        Returns True if written else False if dup val already exists
+        Duplicates are inserted in lexocographic order not insertion order.
+        """
+        return self.addVal(self.vrcs, key, val)
+
+
+    def getVrcs(self, key):
+        """
+        Use dgKey()
+        Return list of receipt triplet at key
+        Returns empty list if no entry at key
+        Duplicates are retrieved in lexocographic order not insertion order.
+        """
+        return self.getVals(self.vrcs, key)
+
+
+    def getVrcsIter(self, key):
+        """
+        Use dgKey()
+        Return iterator of receipt triplets at key
+        Raises StopIteration Error when empty
+        Duplicates are retrieved in lexocographic order not insertion order.
+        """
+        return self.getValsIter(self.vrcs, key)
+
+
+    def cntVrcs(self, key):
+        """
+        Use dgKey()
+        Return count of receipt triplets at key
+        Returns zero if no entry at key
+        """
+        return self.cntVals(self.vrcs, key)
+
+
+    def delVrcs(self, key):
+        """
+        Use dgKey()
+        Deletes all values at key.
+        Returns True If key exists in database Else False
+        """
+        return self.delVals(self.vrcs, key)
+
+
+    def putVres(self, key, vals):
+        """
+        Use dgKey()
+        Write each entry from list of bytes receipt triplets vals to key
+        Adds to existing receipts at key if any
+        Returns True If no error
+        Apparently always returns True (is this how .put works with dupsort=True)
+        Duplicates are inserted in lexocographic order not insertion order.
+        """
+        return self.putVals(self.vres, key, vals)
+
+
+    def addVre(self, key, val):
+        """
+        Use dgKey()
+        Add receipt triplet val bytes as dup to key in db
+        Adds to existing values at key if any
+        Returns True if written else False if dup val already exists
+        Duplicates are inserted in lexocographic order not insertion order.
+        """
+        return self.addVal(self.vres, key, val)
+
+
+    def getVres(self, key):
+        """
+        Use dgKey()
+        Return list of receipt triplet at key
+        Returns empty list if no entry at key
+        Duplicates are retrieved in lexocographic order not insertion order.
+        """
+        return self.getVals(self.vres, key)
+
+
+    def getVresIter(self, key):
+        """
+        Use dgKey()
+        Return iterator of receipt triplets at key
+        Raises StopIteration Error when empty
+        Duplicates are retrieved in lexocographic order not insertion order.
+        """
+        return self.getValsIter(self.vres, key)
+
+
+    def cntVres(self, key):
+        """
+        Use dgKey()
+        Return count of receipt triplets at key
+        Returns zero if no entry at key
+        """
+        return self.cntVals(self.vres, key)
+
+
+    def delVres(self, key):
+        """
+        Use dgKey()
+        Deletes all values at key.
+        Returns True If key exists in database Else False
+        """
+        return self.delVals(self.vres, key)
 
 
     def putKes(self, key, vals):
         """
+        Use snKey()
         Write each key event dig entry from list of bytes vals to key
         Adds to existing event indexes at key if any
         Returns True If at least one of vals is added as dup, False otherwise
@@ -997,6 +1298,7 @@ class Logger(Databaser):
 
     def addKe(self, key, val):
         """
+        Use snKey()
         Add key event val bytes as dup to key in db
         Adds to existing event indexes at key if any
         Returns True if written else False if dup val already exists
@@ -1007,6 +1309,7 @@ class Logger(Databaser):
 
     def getKes(self, key):
         """
+        Use snKey()
         Return list of key event dig vals at key
         Returns empty list if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1016,6 +1319,7 @@ class Logger(Databaser):
 
     def getKeLast(self, key):
         """
+        Use snKey()
         Return last inserted dup key event dig vals at key
         Returns None if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1025,6 +1329,7 @@ class Logger(Databaser):
 
     def cntKes(self, key):
         """
+        Use snKey()
         Return count of dup key event dig val at key
         Returns zero if no entry at key
         """
@@ -1033,6 +1338,7 @@ class Logger(Databaser):
 
     def delKes(self, key):
         """
+        Use snKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
@@ -1082,6 +1388,7 @@ class Logger(Databaser):
 
     def putPses(self, key, vals):
         """
+        Use snKey()
         Write each partial signed escrow event entry from list of bytes dig vals to key
         Adds to existing event indexes at key if any
         Returns True If at least one of vals is added as dup, False otherwise
@@ -1092,6 +1399,7 @@ class Logger(Databaser):
 
     def addPse(self, key, val):
         """
+        Use snKey()
         Add Partial signed escrow val bytes as dup to key in db
         Adds to existing event indexes at key if any
         Returns True if written else False if dup val already exists
@@ -1102,6 +1410,7 @@ class Logger(Databaser):
 
     def getPses(self, key):
         """
+        Use snKey()
         Return list of partial signed escrowed event dig vals at key
         Returns empty list if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1111,6 +1420,7 @@ class Logger(Databaser):
 
     def getPsesLast(self, key):
         """
+        Use snKey()
         Return last inserted dup partial signed escrowed event dig val at key
         Returns None if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1120,6 +1430,7 @@ class Logger(Databaser):
 
     def cntPses(self, key):
         """
+        Use snKey()
         Return count of dup event dig vals at key
         Returns zero if no entry at key
         """
@@ -1128,6 +1439,7 @@ class Logger(Databaser):
 
     def delPses(self, key):
         """
+        Use snKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
@@ -1136,6 +1448,7 @@ class Logger(Databaser):
 
     def putOoes(self, key, vals):
         """
+        Use snKey()
         Write each out of order escrow event dig entry from list of bytes vals to key
         Adds to existing event indexes at key if any
         Returns True If at least one of vals is added as dup, False otherwise
@@ -1146,6 +1459,7 @@ class Logger(Databaser):
 
     def addOoe(self, key, val):
         """
+        Use snKey()
         Add out of order escrow val bytes as dup to key in db
         Adds to existing event indexes at key if any
         Returns True if written else False if dup val already exists
@@ -1156,6 +1470,7 @@ class Logger(Databaser):
 
     def getOoes(self, key):
         """
+        Use snKey()
         Return list of out of order escrow event dig vals at key
         Returns empty list if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1165,6 +1480,7 @@ class Logger(Databaser):
 
     def getOoesLast(self, key):
         """
+        Use snKey()
         Return last inserted dup out of order escrow event dig at key
         Returns None if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1174,6 +1490,7 @@ class Logger(Databaser):
 
     def cntOoes(self, key):
         """
+        Use snKey()
         Return count of dup event dig at key
         Returns zero if no entry at key
         """
@@ -1182,6 +1499,7 @@ class Logger(Databaser):
 
     def delOoes(self, key):
         """
+        Use snKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
@@ -1190,6 +1508,7 @@ class Logger(Databaser):
 
     def putDes(self, key, vals):
         """
+        Use snKey()
         Write each duplicitous event entry dig from list of bytes vals to key
         Adds to existing event indexes at key if any
         Returns True If at least one of vals is added as dup, False otherwise
@@ -1200,6 +1519,7 @@ class Logger(Databaser):
 
     def addDe(self, key, val):
         """
+        Use snKey()
         Add duplicate event index val bytes as dup to key in db
         Adds to existing event indexes at key if any
         Returns True if written else False if dup val already exists
@@ -1210,6 +1530,7 @@ class Logger(Databaser):
 
     def getDes(self, key):
         """
+        Use snKey()
         Return list of duplicitous event dig vals at key
         Returns empty list if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1219,6 +1540,7 @@ class Logger(Databaser):
 
     def getDesLast(self, key):
         """
+        Use snKey()
         Return last inserted dup duplicitous event dig vals at key
         Returns None if no entry at key
 
@@ -1229,6 +1551,7 @@ class Logger(Databaser):
 
     def cntDes(self, key):
         """
+        Use snKey()
         Return count of dup event dig vals at key
         Returns zero if no entry at key
         """
@@ -1237,6 +1560,7 @@ class Logger(Databaser):
 
     def delDes(self, key):
         """
+        Use snKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
@@ -1265,6 +1589,7 @@ class Logger(Databaser):
 
     def putLdes(self, key, vals):
         """
+        Use snKey()
         Write each likely duplicitous event entry dig from list of bytes vals to key
         Adds to existing event indexes at key if any
         Returns True If at least one of vals is added as dup, False otherwise
@@ -1275,6 +1600,7 @@ class Logger(Databaser):
 
     def addLde(self, key, val):
         """
+        Use snKey()
         Add likely duplicitous escrow val bytes as dup to key in db
         Adds to existing event indexes at key if any
         Returns True if written else False if dup val already exists
@@ -1285,6 +1611,7 @@ class Logger(Databaser):
 
     def getLdes(self, key):
         """
+        Use snKey()
         Return list of likely duplicitous event dig vals at key
         Returns empty list if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1294,6 +1621,7 @@ class Logger(Databaser):
 
     def getLdesLast(self, key):
         """
+        Use snKey()
         Return last inserted dup likely duplicitous event dig at key
         Returns None if no entry at key
         Duplicates are retrieved in insertion order.
@@ -1303,6 +1631,7 @@ class Logger(Databaser):
 
     def cntLdes(self, key):
         """
+        Use snKey()
         Return count of dup event dig at key
         Returns zero if no entry at key
         """
@@ -1311,9 +1640,66 @@ class Logger(Databaser):
 
     def delLdes(self, key):
         """
+        Use snKey()
         Deletes all values at key.
         Returns True If key exists in database Else False
         """
         return self.delIoVals(self.ldes, key)
 
 
+class BaserDoer(doing.Doer):
+    """
+    Basic Baser Doer ( LMDB Database )
+
+    Inherited Attributes:
+        .done is Boolean completion state:
+            True means completed
+            Otherwise incomplete. Incompletion maybe due to close or abort.
+
+    Attributes:
+        .baser is Baser or LMDBer subclass
+
+    Inherited Properties:
+        .tyme is float ._tymist.tyme, relative cycle or artificial time
+        .tock is float, desired time in seconds between runs or until next run,
+                 non negative, zero means run asap
+
+    Properties:
+
+    Methods:
+        .wind  injects ._tymist dependency
+        .__call__ makes instance callable
+            Appears as generator function that returns generator
+        .do is generator method that returns generator
+        .enter is enter context action method
+        .recur is recur context action method or generator method
+        .exit is exit context method
+        .close is close context method
+        .abort is abort context method
+
+    Hidden:
+       ._tymist is Tymist instance reference
+       ._tock is hidden attribute for .tock property
+    """
+
+    def __init__(self, baser, **kwa):
+        """
+        Inherited Parameters:
+           tymist is Tymist instance
+           tock is float seconds initial value of .tock
+
+        Parameters:
+           baser is Baser instance
+        """
+        super(BaserDoer, self).__init__(**kwa)
+        self.baser = baser
+
+
+    def enter(self):
+        """"""
+        self.baser.reopen()
+
+
+    def exit(self):
+        """"""
+        self.baser.close()
